# -*- coding: utf-8 -*-
from typing import NamedTuple, Optional, Tuple

import numpy as np
from signalworks import dsp, viterbi
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import Partition, TimeValue, Wave


class ActivityDetector(Processor):
    name = "Activity Detector"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        self.parameters = {
            "threshold": -30.0,
            "smooth": 1.0,
            "frame_size": 0.020,  # in seconds
            "frame_rate": 0.01,
        }

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Partition, TimeValue]:
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        assert isinstance(wav, Wave)
        wav = wav.convert_dtype(np.float64)
        self.progressTracker.update(10)
        assert isinstance(wav, Wave)
        M, time, frequency = dsp.spectrogram(
            wav, self.parameters["frame_size"], self.parameters["frame_rate"]
        )
        self.progressTracker.update(20)
        # Emax = np.atleast_2d(np.max(M, axis=1)).T
        Emax = 20 * np.log10(np.mean((10 ** (M / 10)), axis=1) ** 0.5)
        P = np.empty((len(Emax), 2))
        P[:, 0] = 1 / (1 + np.exp(Emax - self.parameters["threshold"]))
        P[:, 1] = 1 - P[:, 0]  # complement
        self.progressTracker.update(30)
        seq, _ = viterbi.search_smooth(P, self.parameters["smooth"])
        self.progressTracker.update(90)
        tmv = TimeValue(
            time,
            seq,
            wav.fs,
            wav.duration,
            wav.path.with_name(wav.path.stem + "-act").with_suffix(
                TimeValue.default_suffix
            ),
        )
        par = Partition.from_TimeValue(tmv)
        par.value = np.char.mod("%d", par.value)
        emax = TimeValue(
            time,
            Emax,
            wav.fs,
            wav.duration,
            wav.path.with_name(wav.path.stem + "-emax").with_suffix(
                TimeValue.default_suffix
            ),
        )
        emax.min = Emax.min()
        emax.max = Emax.max()
        emax.unit = "dB"
        emax.label = "maximum frequency magnitude"
        return par, emax



# -*- coding: utf-8 -*-
from typing import Dict, NamedTuple, Optional, Tuple

import numpy as np
from scipy import signal
from signalworks import dsp, viterbi
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import Partition, TimeValue, Wave


class F0Analyzer(Processor):
    name = "F0 Analysis"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        self.t0_min = 0
        self.t0_max = 0
        self.parameters = {
            "smooth": 0.01,
            "f0_min": 51,  # in Hertz
            "f0_max": 300,  # in Hertz
            "frame_size": 0.040,  # in seconds
            "frame_rate": 0.010,  # in seconds
            "dop threshold": 0.7,
            "energy threshold": 0.1,
        }

    def set_parameters(self, parameter: Dict[str, str]) -> None:
        super().set_parameters(parameter)
        assert (
            self.parameters["f0_min"] < self.parameters["f0_max"]
        ), "f0_min must be < f0_max"
        assert self.parameters["frame_size"] > (
            2 / self.parameters["f0_min"]
        ), "frame_size must be > 2 / f0_min"

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[TimeValue, TimeValue, Partition]:
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        assert isinstance(wav, Wave)
        wav = wav.convert_dtype(np.float64)
        self.progressTracker.update(10)
        assert isinstance(wav, Wave)
        R, time, frequency = dsp.correlogram(
            wav, self.parameters["frame_size"], self.parameters["frame_rate"]
        )

        self.progressTracker.update(30)
        assert isinstance(wav, Wave)
        t0_min = int(round(wav.fs / self.parameters["f0_max"]))
        t0_max = int(round(wav.fs / self.parameters["f0_min"]))
        index = np.arange(t0_min, t0_max + 1, dtype=np.int)
        E = R[:, 0]  # energy
        R = R[:, index]  # only look at valid candidates
        # normalize
        R -= R.min()
        R /= R.max()
        # find best sequence
        seq, _ = viterbi.search_smooth(R, self.parameters["smooth"])
        self.progressTracker.update(80)

        f0 = wav.fs / (t0_min + seq)
        # degree of periodicity
        dop = R[np.arange(R.shape[0]), seq]
        # voicing
        v = (
            (dop > self.parameters["dop threshold"])
            & (E > self.parameters["energy threshold"])
            #  (seq > 0) & (seq < len(index) - 1)
        ).astype(np.int)
        v = signal.medfilt(v, 5)  # TODO: replace by a 2-state HMM
        f0[v == 0] = np.nan
        # prepare tracks
        f0 = TimeValue(
            time,
            f0,
            wav.fs,
            wav.duration,
            wav.path.with_name(wav.path.stem + "-f0").with_suffix(
                TimeValue.default_suffix
            ),
        )
        f0.min = self.parameters["f0_min"]
        f0.max = self.parameters["f0_max"]
        f0.unit = "Hz"
        f0.label = "F0"
        dop = TimeValue(
            time,
            dop,
            wav.fs,
            wav.duration,
            wav.path.with_name(wav.path.stem + "-dop").with_suffix(
                TimeValue.default_suffix
            ),
        )
        dop.min = 0
        dop.max = 1
        dop.label = "degree of periodicity"
        vox = TimeValue(
            time,
            v,
            wav.fs,
            wav.duration,
            wav.path.with_name(wav.path.stem + "-vox").with_suffix(
                TimeValue.default_suffix
            ),
        )
        vox = Partition.from_TimeValue(vox)
        vox.label = "voicing"
        assert isinstance(f0, TimeValue)
        assert isinstance(dop, TimeValue)
        assert isinstance(vox, Partition)
        return f0, dop, vox



# -*- coding: utf-8 -*-
from typing import Dict, NamedTuple, Optional, Tuple

import numpy as np
from signalworks import dsp, viterbi
from signalworks.processors.processing import (
    DefaultProgressTracker,
    InvalidParameterError,
    Processor,
)
from signalworks.tracking import Partition, TimeValue, Wave


class PeakTracker(Processor):
    name = "Peak Tracker"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        self.parameters = {
            "smooth": 1.0,
            "freq_min": 100,
            "freq_max": 1000,
            "frame_size": 0.02,  # seconds, determines freq res.
            "frame_rate": 0.01,
            "NFFT": 512,
        }

    def get_parameters(self):
        if "wave" in self.acquire._field_types.keys():
            if self.data.wave is not None:
                self.parameters["freq_max"] = self.data.wave.fs / 2
        return super().get_parameters()

    def set_parameters(self, parameter: Dict[str, str]) -> None:
        super().set_parameters(parameter)
        if not self.parameters["freq_min"] < self.parameters["freq_max"]:
            raise InvalidParameterError("freq_min must be < freq_max")

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[TimeValue]:
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        self.progressTracker.update(10)
        assert isinstance(wav, Wave)
        ftr, time, frequency = dsp.spectrogram(
            wav,
            self.parameters["frame_size"],
            self.parameters["frame_rate"],
            NFFT=self.parameters["NFFT"],
        )
        self.progressTracker.update(50)
        a = frequency.searchsorted(self.parameters["freq_min"])
        b = frequency.searchsorted(self.parameters["freq_max"])
        # import time as timer
        # print('searching')
        # tic = timer.time()
        seq, _ = viterbi.search_smooth(ftr[:, a:b], self.parameters["smooth"])
        self.progressTracker.update(90)
        # toc = timer.time()
        # print(f'done, took: {toc-tic}')
        trk = TimeValue(
            time,
            frequency[a + seq],
            wav.fs,
            wav.duration,
            wav.path.with_name(wav.path.stem + "-peak").with_suffix(
                TimeValue.default_suffix
            ),
        )
        trk.min = 0
        trk.max = wav.fs / 2
        trk.unit = "Hz"
        trk.label = "frequency"
        return (trk,)


class PeakTrackerActiveOnly(PeakTracker):
    name = "Peak Tracker (active regions only)"
    acquire = NamedTuple("acquire", [("wave", Wave), ("active", Partition)])

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[TimeValue]:
        if progressTracker is not None:
            self.progressTracker = progressTracker
        peak = super().process()[0]
        active = self.data.active
        assert isinstance(active, Partition)
        for i in range(len(active.time) - 1):
            if active.value[i] in ["0", 0]:
                a = np.searchsorted(peak.time / peak.fs, active.time[i] / active.fs)
                b = np.searchsorted(peak.time / peak.fs, active.time[i + 1] / active.fs)
                peak.value[a:b] = np.nan
        return (peak,)



# -*- coding: utf-8 -*-
from pathlib import Path

import pytest  # noqa: F401
import signalworks
from signalworks.tracking import Track


@pytest.fixture(scope="function")
def speech_track():
    speech_wav = Path(signalworks.__file__).parents[1] / "data" / "speech-mwm.wav"
    track = Track.read(speech_wav)
    return track



import logging
import os
from pathlib import Path
from typing import Optional

import numpy
from scipy.io.wavfile import read as wav_read
from scipy.io.wavfile import write as wav_write
from scipy.signal import resample_poly
from signalworks.tracking.error import MultiChannelError
from signalworks.tracking.metatrack import MetaTrack

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class Wave(MetaTrack):
    """monaural waveform"""

    default_suffix = ".wav"

    def __init__(
        self,
        value: numpy.ndarray,
        fs: int,
        duration: int = None,
        offset: int = 0,
        path: str = None,
    ) -> None:
        super().__init__()
        if path is None:
            path = str(id(self))
        self.min: Optional[int] = None
        self.max: Optional[int] = None
        # TODO: what happens if path is None
        self.path = Path(path).with_suffix(self.default_suffix)
        assert isinstance(value, numpy.ndarray)
        assert 1 <= value.ndim, "only a single channel is supported"
        assert isinstance(fs, int)
        assert fs > 0
        self._value = value
        self._fs = fs
        self._offset = (
            offset
        )  # this is required to support heterogenous fs in multitracks
        self.type = "Wave"
        self.label = f"amplitude-{value.dtype}"
        if not duration:
            duration = len(self._value)
        assert len(self._value) <= duration < len(self._value) + 1, (
            "Cannot set duration of a wave to other than a number in "
            "[length, length+1), where length = len(self.value)"
        )
        self._duration = duration

    def get_offset(self):
        return self._offset

    def set_offset(self, offset):
        assert 0.0 <= offset < 1.0
        self._offset = offset

    offset = property(get_offset, set_offset)

    def get_time(self):
        return (
            numpy.arange(len(self._value))
            if self._offset == 0
            else numpy.arange(len(self._value)) + self._offset
        )

    def set_time(self, time):
        raise Exception("can't set times for Wave")

    time = property(get_time, set_time)

    def get_value(self):
        return self._value

    def set_value(self, value):
        assert isinstance(value, numpy.ndarray)
        assert 1 == value.ndim, "only a single channel is supported"
        self._value = value
        if not (len(self._value) <= self._duration < len(self._value) + 1):
            self._duration = len(self._value)

    value = property(get_value, set_value)

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def get_duration(self):
        return self._duration

    def set_duration(self, duration):
        assert len(self._value) <= duration < len(self._value) + 1, (
            "Cannot set duration of a wave to other than a number in [length, length+1) "
            "- where length = len(self.value)"
        )
        self._duration = duration

    duration = property(get_duration, set_duration)

    #  def get_channels(self):
    #      return 1 if self._value.ndim == 1 else self._value.shape[1] # by convention

    #  def set_channels(self, *args):
    #      raise Exception("Cannot change wave channels - create new wave instead")

    #  channels = property(get_channels)  #, set_channels)

    def get_dtype(self):
        return self._value.dtype

    #  def set_dtype(self, *args):
    #      raise Exception("Cannot change wave dtype - create new wave instead")
    dtype = property(get_dtype)

    def get_bitdepth(self):
        dtype = self._value.dtype
        if dtype == numpy.int16:
            return 16
        elif dtype == numpy.float32 or dtype == numpy.int32:
            return 32
        elif dtype == numpy.float64:
            return 64
        else:
            raise Exception("unknown dtype = %s" % dtype)

    bitdepth = property(get_bitdepth)

    def __eq__(self, other):
        if (
            (self._fs == other._fs)
            and (self._duration == other._duration)
            and (len(self._value) == len(other._value))
            and (self._value == other._value).all()
        ):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "\n".join(
            [
                f"value={self.value}",
                f"min={self.value.min()}",
                f"max={self.value.max()}",
                f"dtype={self.dtype}",
                f"fs={self.fs}",
                f"duration={self.duration}",
            ]
        )

    def __len__(self):
        return len(self._value)

    def __add__(self, other):
        assert type(other) == type(self), "Cannot add Track objects of different types"
        assert self._fs == other._fs, "sampling frequencies must match"
        assert self.value.dtype == other.value.dtype, "dtypes must match"
        value = numpy.concatenate((self._value, other._value))
        return type(self)(value, self.fs)

    def resample(self, fs):
        """resample to a certain fs"""
        assert isinstance(fs, int)
        if fs != self._fs:
            if fs > self._fs:  # upsampling
                return type(self)(resample_poly(self._value, int(fs / self._fs), 1), fs)
            else:  # downsampling
                return type(self)(resample_poly(self._value, 1, int(self._fs / fs)), fs)
        else:
            return self

    def _convert_dtype(self, source, target_dtype):
        """
        return a link (if unchanged) or copy of signal in the specified dtype (often changes bit-depth as well)
        """
        assert isinstance(source, numpy.ndarray)
        source_dtype = source.dtype
        assert source_dtype in (
            numpy.int16,
            numpy.int32,
            numpy.float32,
            numpy.float64,
        ), "source must be a supported type"
        assert target_dtype in (
            numpy.int16,
            numpy.int32,
            numpy.float32,
            numpy.float64,
        ), "target must be a supported type"
        if source_dtype == target_dtype:
            return source
        else:  # conversion
            if source_dtype == numpy.int16:
                if target_dtype == numpy.int32:
                    return source.astype(target_dtype) << 16
                else:  # target_dtype == numpy.float32 / numpy.float64:
                    return source.astype(target_dtype) / (1 << 15)
            elif source_dtype == numpy.int32:
                if target_dtype == numpy.int16:
                    return (source >> 16).astype(target_dtype)  # lossy
                else:  # target_dtype == numpy.float32 / numpy.float64:
                    return source.astype(target_dtype) / (1 << 31)
            else:  # source_dtype == numpy.float32 / numpy.float64
                M = numpy.max(numpy.abs(source))
                limit = 1 - 1e-16
                if M > limit:
                    factor = limit / M
                    logger.warning(
                        f"maximum float waveform value {M} is beyond [-{limit}, {limit}],"
                        f"applying scaling of {factor}"
                    )
                    source *= factor
                if target_dtype == numpy.float32 or target_dtype == numpy.float64:
                    return source.astype(target_dtype)
                else:
                    if target_dtype == numpy.int16:
                        return (source * (1 << 15)).astype(target_dtype)  # dither?
                    else:  # target_dtype == numpy.int32
                        return (source * (1 << 31)).astype(target_dtype)  # dither?

    def convert_dtype(self, target_dtype):
        """returns a new wave with the waveform in the specified target_dtype"""
        # TODO: take care of setting new min and max
        return type(self)(
            self._convert_dtype(self._value, target_dtype), self._fs, path=self.path
        )

    def select(self, a, b):
        assert a >= 0
        assert a < b  # or a <= b?
        assert b <= self.duration
        # TODO: modify this for float a and b
        return type(self)(self._value[a:b], self._fs)

    @classmethod
    def read(cls, name, *_args, **_kwargs):
        """Loads object from name, adding default extension if missing."""
        name_wo_ext = os.path.splitext(name)[0]
        if name == name_wo_ext:
            name += cls.default_suffix
        return cls.read_wav(name)

    @classmethod
    def read_wav(cls, path, channel=None, mmap=False):
        """load waveform from file"""
        fs, value = wav_read(path, mmap=mmap)
        if value.ndim == 1:
            if channel is not None and channel != 0:
                raise MultiChannelError(
                    "cannot select channel {} from monaural file {}".format(
                        channel, path
                    )
                )
        if value.ndim == 2:
            if channel is None:
                raise MultiChannelError(
                    f"must select channel when loading file {path} with {value.shape[1]} channels"
                )
            try:
                value = value[:, channel]
            except IndexError:
                raise MultiChannelError(
                    f"cannot select channel {channel} from file "
                    f"{path} with {value.shape[1]} channels"
                )
        wav = Wave(value, fs, path=path)
        if value.dtype == numpy.int16:
            wav.min = -32767
            wav.max = 32768
        else:
            raise NotImplementedError
        return wav

    wav_read = read_wav

    def write(self, name, *_args, **_kwargs):
        """Saves object to name, adding default extension if missing."""
        name_wo_ext = os.path.splitext(name)[0]
        if name == name_wo_ext:
            name += self.default_suffix
        self.write_wav(name)

    def write_wav(self, name):
        """save waveform to file
        The bits-per-sample will be determined by the data-type (mostly)"""
        wav_write(name, self._fs, self._value)

    wav_write = write_wav

    def __getitem__(self, index):
        return Wave(self._value[index], self.fs)

    # def __setitem__(self, index, value):
    #    self._value[index] = value

    # def __add__(self, other):
    #     """wave1 + wave2"""
    #     if self.fs != other.fs:
    #         raise Exception("sampling frequency of waves must match")
    #     return type(self)(numpy.concatenate((self.va, other.va)), self.fs)  # return correct (child) class

    # def delete(self, a, b, fade = 0):
    #     pass

    # def cut(self, a, b, fade = 0):
    #     wave = self.copy(a, b)
    #     self.delete(a, b, fade)
    #     return wave

    # def insert(self, wave, a, fade = 0):
    #     """insert wave into self at time a"""
    #     if self.fs != wave.fs:
    #         raise Exception("sampling frequency of waves must match")
    #     if fade:
    #         n = round(fade * self.fs)
    #         if n*2 > len(wave.signal):
    #             raise Exception("fade inverval is too large")
    #         up = numpy.linspace(0, 1, n)
    #         down = numpy.linspace(1, 0, n)
    #         p = wave.signal.copy()
    #         p[:n] *= up
    #         p[-n:] *= down
    #         l = self.signal[:a+n]
    #         l[-n:] *= down
    #         r = self.signal[a-n:]
    #         r[:n] *= up
    #     else:
    #         self.signal = numpy.concatenate((self.signal[:a], wave.signal, self.signal[a:]))

    def crossfade(self, wave, length):
        """append wave to self, using a crossfade of a specified length in samples"""
        assert type(self) == type(wave), "Cannot add Track objects of different types"
        assert self.fs == wave.fs, "sampling frequency of waves must match"
        assert isinstance(length, int)
        assert length > 0
        assert wave.duration >= length
        assert self.duration >= length
        ramp = numpy.linspace(1, 0, length + 2)[1:-1]  # don't include 0 and 1
        value = self.value.copy()
        value[-length:] = value[-length:] * ramp + wave.value[:length] * (
            1 - ramp
        )  # TODO: think about dtypes here
        value = numpy.concatenate((value, wave.value[length:]))
        return type(self)(value, self.fs)

    # TODO: Test / fix me!
    def time_warp(self, x, y):
        raise NotImplementedError
        logger.warning(
            "time_warping wave, most of the time this is not what is desired"
        )
        time = numpy.arange(len(self._value))
        # time = index / self._fs
        time = numpy.round(numpy.interp(time, x, y)).astype(numpy.int)
        # index = int(time * self.fs)
        self._value = self._value[time]



# -*- coding: utf-8 -*-
from typing import NamedTuple, Optional, Tuple

import numpy as np
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import Wave

# @abstractmethod
# def process(self, progressTracker: Optional[DefaultProgressTracker] = None) -> Tuple[Tracks, ...]:


class ConverterToFloat64(Processor):
    name = "Conversion to Float64"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Wave]:
        _ = Processor.process(self)
        wav = self.data.wave
        assert isinstance(wav, Wave)
        wav = wav.convert_dtype(np.float64)
        assert isinstance(wav, Wave)
        wav.path = wav.path.with_name(wav.path.stem + "-float64").with_suffix(
            Wave.default_suffix
        )
        return (wav,)


class ConverterToInt16(Processor):
    name = "Conversion to Int16"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Wave]:  # , **kwargs
        Processor.process(self)
        wav = self.data.wave
        assert isinstance(wav, Wave)
        wav = wav.convert_dtype(np.int16)
        assert isinstance(wav, Wave)
        wav.path = wav.path.with_name(wav.path.stem + "-int16").with_suffix(
            Wave.default_suffix
        )
        return (wav,)



# -*- coding: utf-8 -*-
from typing import NamedTuple, Optional, Tuple

import numpy as np
from signalworks import dsp
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import Wave


class NoiseReducer(Processor):
    name = "Noise Reducer"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        self.parameters = {"silence_percentage": 10, "frame_rate": 0.01}  # in seconds

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[Wave]:
        # Processor.process(self, **kwargs)
        if progressTracker is not None:
            self.progressTracker = progressTracker
        inp = self.data.wave
        assert isinstance(inp, Wave)
        inp = inp.convert_dtype(np.float64)
        self.progressTracker.update(20)
        # TODO: pull this up into here
        assert isinstance(inp, Wave)
        out = dsp.spectral_subtract(
            inp, self.parameters["frame_rate"], self.parameters["silence_percentage"]
        )
        self.progressTracker.update(90)
        assert isinstance(inp, Wave)
        out.path = inp.path.with_name(inp.path.stem + "-denoised").with_suffix(
            Wave.default_suffix
        )
        return (out,)



import logging
import os
from collections import Iterable

import numpy
from signalworks.tracking.metatrack import MetaTrack
from signalworks.tracking.partition import Partition

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

TIME_TYPE = numpy.int64


class TimeValue(MetaTrack):
    def __init__(self, time, value, fs, duration, path=None):
        super().__init__()
        assert isinstance(time, numpy.ndarray)
        assert time.ndim == 1
        assert time.dtype == TIME_TYPE
        assert (
            numpy.diff(time.astype(numpy.float)) > 0
        ).all(), "times must be strictly monotonically increasing"
        assert isinstance(value, numpy.ndarray)
        assert isinstance(fs, int)
        assert fs > 0
        assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        assert len(time) == len(value), "length of time and value must match"
        assert not (len(time) and duration <= time[-1]), "duration is not > times"
        self._time = time
        self._value: numpy.ndarray = value
        self._fs = fs
        self._duration = duration
        self.min = numpy.nanmin(value)
        self.max = numpy.nanmax(value)
        self.unit = ""
        self.label = ""
        self.path = path

    def get_time(self):
        assert (
            numpy.diff(self._time.astype(numpy.float)) > 0
        ).all(), (
            "times must be strictly monotonically increasing"
        )  # in case the user messed with .time[index] directly
        return self._time

    def set_time(self, time):
        assert isinstance(time, numpy.ndarray)
        assert time.ndim == 1
        assert time.dtype == TIME_TYPE
        assert not (len(time) and self._duration <= time[-1]), "duration is not > times"
        assert (
            numpy.diff(time.astype(numpy.float)) > 0
        ).all(), "times must be strictly monotonically increasing"
        assert len(time) == len(self._value), "length of time and value must match"
        self._time = time

    time = property(get_time, set_time)

    def get_value(self):
        return self._value

    def set_value(self, value):
        assert isinstance(value, numpy.ndarray)
        assert len(self._time) == len(value), "length of time and value must match"
        self._value = value

    value = property(get_value, set_value)

    def get_duration(self):
        return self._duration

    def set_duration(
        self, duration
    ):  # assume times are available, if not, this must be overridden
        assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        assert not (
            len(self._time) and duration <= self._time[-1]
        ), "duration is not > times"
        self._duration = duration

    duration = property(get_duration, set_duration, doc="duration")

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def __eq__(self, other):
        if (
            (self._fs == other._fs)
            and (self._duration == other._duration)
            and (len(self._time) == len(other._time))
            and (len(self._value) == len(other._value))
            and (self._time == other._time).all()
            and numpy.allclose(
                numpy.round(self._value, 3), numpy.round(other._value, 3)
            )
        ):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self._time)

    def __str__(self):
        return f"{list(zip(self._time, self._value))}, fs={self._fs}, duration={self._duration}, path={self.path}"

    def __add__(self, other):
        assert type(other) == type(self), "Cannot add Track objects of different types"
        assert self.fs == other.fs, "sampling frequencies must match"
        time = numpy.concatenate(
            (self.time, (other.time + self.duration).astype(other.time.dtype))
        )
        value = numpy.concatenate((self.value, other.value))
        duration = self.duration + other.duration
        return type(self)(time, value, self.fs, duration)

    # def __iadd__(self, other):
    #     assert type(other) == type(self), "Cannot add Track objects of different types"
    #     assert self.fs == other.fs, "sampling frequencies must match"
    #     self._time = numpy.concatenate((self._time, (other._time + self._duration).astype(other.time.dtype)))
    #     self._value = numpy.concatenate((self._value, other.value))
    #     duration = self.duration + other.duration
    #     return self

    def resample(self, fs):
        if fs != self._fs:
            factor = fs / self._fs
            time = numpy.round(factor * self._time).astype(TIME_TYPE)
            assert (
                numpy.diff(time) > 0
            ).all(), (
                "new fs causes times to fold onto themselves due to lack in precision"
            )
            # need to use numpy.round for consistency - it's different from the built-in round
            duration = int(numpy.ceil(factor * self._duration))
            return type(self)(time, self._value, fs, duration)
        else:
            return self

    def select(self, a, b):
        """(SHOULD!) return a copy"""
        # assert isinstance(a, TIME_TYPE)  # this doesn't seem necessary
        # assert isinstance(b, TIME_TYPE)
        assert b > a
        ai = self.time.searchsorted(a)
        bi = self.time.searchsorted(b)
        time = self._time[ai:bi] - a  # not possible to make this a view
        value = self._value[ai:bi]  # this is a view - should probably be a copy.
        return type(self)(time, value, self.fs, TIME_TYPE(b - a))

    default_suffix = ".tmv"

    @classmethod
    def read(cls, name, track_name=None, fs=300_000, duration=None, *_args, **_kwargs):
        """Loads object from name, adding default extension if missing."""
        if name == os.path.splitext(name)[0]:
            if track_name:
                ext = "." + track_name
            else:
                ext = cls.default_suffix
            name += ext
        else:
            ext = os.path.splitext(name)[1].lower()
        if ext in (".pit", ".nrg", cls.default_suffix):
            self = cls.read_tmv(name, fs)
        elif ext == ".f0":
            self = cls.read_pitchtier(name, fs)[0]
        elif ext == ".pitchtier":
            self = cls.read_pitchtier(name, fs)
        else:
            self = None
            try:
                if self is None:
                    self = cls.read_f0(name, fs)[0]
            except FileNotFoundError:  # TODO: need more exceptions?
                pass
            try:
                if self is None:
                    self = cls.read_pitchtier(name, fs)
            except FileNotFoundError:  # TODO: need more exceptions?
                pass
            if self is None:
                raise ValueError("file '{}' has unknown format".format(name))
        if duration:
            self.set_duration(duration)
        self.path = name
        return self

    @classmethod
    def read_f0(cls, name, frameRate=0.01, frameSize=0.0075, fs=48000):
        # return TimeValue(numpy.ndarray([1]).astype(TIME_TYPE), numpy.ndarray([1]), fs, 1)
        # 4 fields for each frame, pitch, probability of voicing, local root mean squared measurements,
        # and the peak normalized cross-correlation value frame arguments are in seconds
        f = open(name, "r")
        lines = f.readlines()
        f.close()
        F = len(lines)
        time = numpy.round((numpy.arange(F) * frameRate + frameSize / 2) * fs).astype(
            TIME_TYPE
        )
        duration = numpy.round((F * frameRate + frameSize) * fs).astype(TIME_TYPE)
        # time = numpy.arange(len(lines)) * frameRate + frameSize / 2
        value_f0 = numpy.zeros(F, numpy.float32)
        value_vox = numpy.zeros(F, numpy.float32)
        import re

        form = re.compile(r"^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$")
        for i, line in enumerate(lines):
            match = form.search(line)
            if not match:
                # logger.error('badly formatted f0 file: {}'.format("".join(lines)))
                raise Exception
                # return (TimeValue(numpy.array([0]), numpy.array([100]), fs, duration),
                #         Partition(numpy.array([0, duration]), numpy.array([0]), fs))
            f0, voicing, _energy, _xcorr = match.groups()
            value_f0[i] = float(f0)
            value_vox[i] = float(voicing)
        index = numpy.where(value_f0 > 0)[0]  # keep only nonzeros F0 values
        pit = TimeValue(time[index], value_f0[index], fs, duration)
        vox = Partition.from_TimeValue(
            TimeValue(time, value_vox, fs, duration)
        )  # return a Partition
        return pit, vox

    f0read = read_f0

    @classmethod
    def read_tmv(cls, name, fs=300000):
        obj = numpy.loadtxt(name)
        time = obj[:, 0]
        value = obj[:, 1]
        time = numpy.round(time * fs).astype(TIME_TYPE)
        duration = (time[-1] + 1).astype(TIME_TYPE)
        return TimeValue(time, value, fs, duration, path=name)

    @classmethod
    def read_frm(cls, name, fs=48000):
        f = open(name, "r")
        lines = f.readlines()
        f.close()
        frameRate = 0.01
        frameSize = 0.049
        time = numpy.round(
            (numpy.arange(len(lines)) * frameRate + frameSize / 2) * fs
        ).astype(TIME_TYPE)
        duration = (time[-1] + 1).astype(TIME_TYPE)
        for i, line in enumerate(lines):
            fb = numpy.array(line.split()).astype(numpy.float)
            if i == 0:
                value = numpy.zeros((len(lines), len(fb)), numpy.float64)
            value[i, :] = fb
        return TimeValue(time, value, fs, duration, path=name)

    frmread = read_frm

    def write(self, name, track_name=None, *_args, **_kwargs):
        """Saves object to name, adding default extension if missing."""
        name_wo_ext = os.path.splitext(name)[0]
        if name == name_wo_ext:
            if track_name:
                ext = "." + track_name
            else:
                ext = self.default_suffix
            name += ext
        self.write_tmv(name)

    def write_f0(self, name, frameRate=0.01, frameSize=0.0075, fs=48000):
        raise NotImplementedError("TimeValue.write_f0: not yet implemented")

    def write_tmv(self, name):
        obj = numpy.c_[self._time / self._fs, self._value]
        numpy.savetxt(name, obj)
        # with open(name, 'w') as f:
        #     for tv in zip(self._time / self._fs, self._value):
        #         f.write('\t'.join(str(round(x, 16)) for x in tv) + '\n')

    def write_frm(self, name):
        with open(name, "w") as f:
            if isinstance(self._value[0], Iterable):
                for fb in self._value:
                    f.write("\t".join(str(round(f, 3)) for f in fb) + "\n")
            else:
                for fb in self._value:
                    f.write(str(round(fb, 3)) + "\n")

    @classmethod
    def from_Partition(cls, p):
        """convert a partition track into a time-value track"""
        assert isinstance(p, Partition)
        time = numpy.r_[p.time[:-1], p.time[-1] - 1, p.time[1:-1] - 1]
        time.sort()
        # double this to keep it constant
        value = numpy.array(
            [p.value[int(i)] for i in numpy.arange(0, len(p.value), 0.5)]
        )
        return TimeValue(time, value, p.fs, p.duration)

    # def __getitem__(self, index):  # should I make the default just the value?
    #     return (self._time[index], self._value[index])
    #     return self._value[index] # ??

    # def __setitem__(self, index, value):
    #     self._time[index] = value[0]
    #     self._value[index] = value[1]

    def get_index(self, t):
        """return the index of the nearest available data"""
        if len(self._time) == 0:
            raise Exception
        if t <= 0:
            return 0
        if t >= self._time[-1]:
            return len(self._value) - 1
        # t is within bounds
        ri = self._time.searchsorted(t)
        li = ri - 1
        rt = self._time[ri]
        lt = self._time[li]
        if (t - lt) < (rt - t):
            return li
        else:
            return ri

    def _get_value(self, t, interpolation="nearest"):
        # check bounds
        if len(self._time) == 0:
            raise Exception
        if t < self._time[0]:
            # raise Exception('out of bounds left')
            # logger.warning('out of bounds left')
            v = self._value[0]  # better? yes!
        elif t > self._time[-1]:
            # raise Exception('out of bounds right')
            # logger.warning('out of bounds right')
            v = self._value[-1]  # better? yes!
        else:  # t is within bounds
            ri = self._time.searchsorted(t)  # a numpy function
            rt = self._time[ri]
            rv = self._value[ri]
            if rt == t:  # exact hit
                v = rv
            else:
                li = ri - 1
                lt = self._time[li]
                lv = self._value[li]
                if interpolation == "nearest":
                    if (t - lt) < (rt - t):
                        v = lv
                    else:
                        v = rv
                else:  # linear
                    a = float(t - lt) / float(rt - lt)
                    v = a * rv + (1.0 - a) * lv
        return v

    def get(
        self, T, interpolation="nearest"
    ):  # start using interp_* methods, which are faster
        """return the values at times T"""
        if isinstance(T, numpy.ndarray):
            assert T.ndim == 1
            n = (
                self._value.shape[1] if self._value.ndim > 1 else 1
            )  # this can be done much better I'm sure
            V = numpy.empty((T.shape[0], n), dtype=self._value.dtype)
            for i, t in enumerate(T):
                # TODO: speed this up by using a cached version of the interpolate function
                V[i] = self._get_value(t, interpolation)
            return V
        else:
            return self._get_value(T, interpolation)

    def interp_linear(self, T):
        assert numpy.all(numpy.diff(self.time) > 0)
        return numpy.interp(T, self.time, self.value)

    def interp_nearest(self, T):
        assert numpy.all(numpy.diff(self.time) > 0)
        index = numpy.interp(T, self.time, numpy.arange(T))
        return self.value[numpy.round(index)]  # TODO: test me!

    def interpolate(self, T, kind="linear"):
        """kind : str or int, optional
        Specifies the kind of interpolation as a string
        ('linear', 'nearest', 'zero', 'slinear', 'quadratic, 'cubic'
        where 'slinear', 'quadratic' and 'cubic' refer to a spline
        interpolation of first, second or third order) or as an integer
        specifying the order of the spline interpolator to use.
        Default is 'linear'."""
        from scipy.interpolate import interp1d

        if len(self.time):
            f = interp1d(
                self.time,
                self.value,
                kind=kind,
                axis=0,
                copy=True,
                bounds_error=True,
                fill_value=numpy.nan,
            )
            return f(T)
        else:
            logger.warning("interpolating without data")
            return numpy.ones(len(T)) * numpy.nan

    # def select_index(self, a, b):
    #     """return indeces such that a <= time < b"""
    #     assert b > a
    #     return range(self.time.searchsorted(a), self.time.searchsorted(b))

    # def select(self, a, b):
    #     """copy a section of the track, interval [)"""
    #     #indexa = numpy.where((self.ti - a) >= 0)[0][0] # first inside
    #     #indexb = numpy.where((self.ti - b) <= 0)[0][-1] # last inside
    #     index = self.select_index(a, b)
    #     if len(index) == 0:
    #         if type(self._value) == numpy.ndarray:
    #             value = numpy.array([])
    #         else:
    #             value = []
    #         return type(self)(time=numpy.empty(0, dtype=TIME_TYPE), value=value, fs=self.fs, duration=b-a)
    #     # on the line above, I first had [0], which cause all subsequent += to be in int, causing a bad bug.
    #     # Perhaps this is the reason to choose all times as int
    #     ai = index[0]
    #     bi = index[-1]
    #     time = self._time[ai:bi] #.copy() - don't think that's needed here ...
    #     value = self._value[ai:bi] #.copy()
    #     ## now, to copy the track accurately, we have to go all the way to the edges, and capture the values there,
    #     ## in accordance to the underlying interpolation function
    #     ### limit selection
    #     ##if a < self._time[0]:
    #         ##a = self.ti[0]
    #     ##if b > self.ti[-1]:
    #         ##b = self.ti[-1]
    #     ###if a < self._time[index[0]]:  # prepend - not sure if this is needed
    #         ###time  = numpy.concatenate(([a], time))
    #         ###value = numpy.concatenate(([self.get(a, interpolation)], value))
    #     ## can't append in any way that make sense, I _think_ TODO: Check this
    #     ##if self.time[index[-1]] <= b:
    #         ##time = numpy.concatenate((time, [b]))
    #         ##value = numpy.concatenate((value, f(b).T))
    #     ##time -= time[0]
    #     time -= a # TODO: Check this!
    #     tv = type(self)(time=time, value=value, fs=self.fs, duration=b-a)
    #     assert type(self._value) == type(tv.value)
    #     return tv

    def time_warp(self, X, Y):
        assert X[0] == 0
        assert Y[0] == 0
        assert X[-1] == self.duration
        time = numpy.interp(self.time, X, Y).astype(self.time.dtype)
        if 0:
            from matplotlib import pylab

            pylab.plot(X, Y, "rx-")
            for x, y in zip(self.time, time):
                pylab.plot([x, x, 0], [0, y, y])
            pylab.show()
        assert len(numpy.where(numpy.diff(time) <= 0)[0]) == 0  # no collapsed items
        self._time = time  # don't do duration check yet
        self.duration = Y[-1]  # changed this from _duration

    def interp(self, time):  # TODO: unify with get()
        if self._value.ndim == 1:
            return numpy.interp(
                time, self._time, self._value, self._value[0], self._value[-1]
            )
        elif self._value.ndim == 2:
            value = numpy.empty(
                (len(time), self._value.shape[1]), dtype=self._value.dtype
            )
            for j in range(self._value.shape[1]):
                value[:, j] = numpy.interp(
                    time,
                    self._time,
                    self._value[:, j],
                    self._value[0, j],
                    self._value[-1, j],
                )
            return value
        else:
            raise Exception



import logging
from pathlib import Path

import numpy
from signalworks.tracking.error import LabreadError
from signalworks.tracking.metatrack import MetaTrack

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

TIME_TYPE = numpy.int64


class Label(MetaTrack):
    """Like Partition, but label regions do NOT have to be contiguous"""

    default_suffix = ".lab"

    def check(self):
        """value[k] has beginning and ending times time[2*k] and time[2*k+1]"""
        assert isinstance(self._time, numpy.ndarray)
        assert self._time.ndim == 1
        assert self._time.dtype == TIME_TYPE
        # assert (numpy.diff(self._time.astype(numpy.float)) >= 0).all(),\
        # "times must be (non-strictly) monotonically increasing"
        if not (numpy.diff(self._time.astype(numpy.float)) >= 0).all():
            logger.warning(
                "Label times must be (non-strictly) monotonically increasing"
            )
        assert isinstance(self._value, numpy.ndarray)
        # assert self._value.ndim == 1 # TODO: can I remove this?
        assert isinstance(self._fs, int)
        assert self._fs > 0
        assert (
            numpy.diff(self._time[::2]) > 0
        ).all(), (
            "zero-duration labels are not permitted (but abutting labels are permitted)"
        )
        # this means an empty partition contains one time value at 0!!!
        assert len(self._time) == 2 * len(
            self._value
        ), "length of time and value *2 must match"
        assert isinstance(self._duration, TIME_TYPE) or isinstance(self._duration, int)
        assert not (
            len(self._time) and self._duration < self._time[-1]
        ), "duration is not >= times"
        return True

    def __init__(self, time, value, fs, duration, path=None):
        super().__init__()
        if path is None:
            path = str(id(self))
        self.path = Path(path).with_suffix(self.default_suffix)
        self._time = time
        self._value = value
        self._fs = fs
        self._duration = duration
        self.type = "Label"
        assert self.check()

    def get_time(self):
        if __debug__:
            self.check()
        return self._time

    def set_time(self, time):
        self._time = time
        if __debug__:
            self.check()

    time = property(get_time, set_time)

    def get_value(self):
        if __debug__:
            self.check()
        return self._value

    def set_value(self, value):
        self._value = value
        if __debug__:
            self.check()

    value = property(get_value, set_value)

    def get_duration(self):
        return self._duration

    def set_duration(self, duration):
        assert isinstance(duration, (TIME_TYPE, int, numpy.int64))
        if len(self.time) > 0:
            assert duration > self.time[-1]
        self._duration = duration

    duration = property(get_duration, set_duration, doc="duration of track")

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def __str__(self):
        # s = [u"0"]
        # for i in range(len(self._value)):
        #    s.append(':%s:%i/%.3f' % (self._value[i], self._time[i + 1], self._time[i + 1] / self._fs))
        s = [""]
        for i in range(len(self._value)):
            s.append(
                f"#{i}: %i/%.3f: %s :%i/%.3f\n"
                % (
                    self._time[2 * i],
                    self._time[2 * i] / self._fs,
                    self._value[i],
                    self._time[2 * i + 1],
                    self._time[2 * i + 1] / self._fs,
                )
            )
        s_ = "".join(s)
        return "%sfs=%i\nduration=%i" % (s_, self._fs, self._duration)

    def __add__(self, other):
        if self._fs != other._fs:
            raise Exception("sampling frequencies must match")
        time = numpy.hstack(
            (self._time, self.duration + other._time)
        )  # other._time[0] == 0
        value = numpy.hstack((self._value, other._value))
        duration = self.duration + other.duration
        return Label(time, value, self.fs, duration)

    def __len__(self):
        return len(self._value)

    def resample(self, fs):
        """resample to a certain fs"""
        if fs != self._fs:
            factor = fs / self._fs
            time = numpy.round(factor * self._time).astype(TIME_TYPE)
            time[-1] = numpy.ceil(factor * self._time[-1])
            assert (
                numpy.diff(time) > 0
            ).all(), (
                "new fs causes times to fold onto themselves due to lack in precision"
            )
            duration = numpy.round(factor * self._duration).astype(TIME_TYPE)
            return type(self)(time, self._value, fs, duration)
        else:
            return self

    def select(self, a, b):
        assert 0 <= a
        assert a < b
        assert b <= self.duration
        if __debug__:  # consistency check
            self.check()
        ai = self._time.searchsorted(a)
        bi = self._time.searchsorted(b)
        if ai % 2 == 1:  # inside label
            ai -= 1
        if bi % 2 == 1:  # inside label
            bi += 1
        value = self._value[(ai // 2) : (bi // 2)]
        time = self.time[ai:bi]
        time = time - a
        duration = b - a
        if time[0] < 0:
            time[0] = 0
        if time[-1] > duration:
            time[-1] = duration
        return type(self)(time.astype(TIME_TYPE), value, self.fs, duration)

    @classmethod
    def read_lbl(cls, path, fs=300000, encoding="UTF8"):
        """load times, values from a .lbl label file"""
        with open(path, "r", encoding=encoding) as f:
            lines = f.readlines()
        if len(lines) == 0:
            raise ValueError("file '{}' is empty".format(path))
        lines += "\n"  # make sure it's terminated
        time = []
        value = []
        # label_type = numpy.float64
        for i, line in enumerate(lines):
            if line != "\n":
                try:
                    t1_, t2_, label = line.split()
                except ValueError:
                    logger.warning(
                        'ignoring line "%s" in file %s at line %i' % (line, path, i + 1)
                    )
                    continue
                t1 = float(t1_)  # / 1000  # this particular
                t2 = float(t2_)  # / 1000  # file format
                if label[-1] == "\r":
                    label = label[:-1]
                dur = t2 - t1
                if dur > 0:
                    time.extend([t1, t2])
                    value.append(label)
                elif dur == 0:
                    logger.warning(
                        'zero-length label "%s" in file %s at line %i, ignoring'
                        % (label, path, i + 1)
                    )
                else:
                    raise LabreadError(
                        "label file contains times that are not monotonically increasing"
                    )
        if len(time) == 0:
            raise (Exception("file is empty or all lines were ignored"))
        time = numpy.round(numpy.array(time) * fs).astype(TIME_TYPE)
        # assert labels are not longer than X characters
        value = numpy.array(value)
        # best guess at duration (+1 sec)
        lab = Label(time, value, fs=fs, duration=TIME_TYPE(time[-1] + fs), path=path)
        return lab

    @classmethod
    def read(cls, *args, **kwargs):
        return cls.read_lbl(cls, *args, **kwargs)



import contextlib
import logging
import os
from pathlib import Path

import numpy
from signalworks.tracking.metatrack import MetaTrack

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

TIME_TYPE = numpy.int64


class Event(MetaTrack):
    def __init__(self, time, fs, duration, path=None):
        super().__init__()
        if path is None:
            path = str(id(self))
        self.path = Path(path).with_suffix(".trk")
        assert isinstance(time, numpy.ndarray)
        assert time.ndim == 1
        assert time.dtype == TIME_TYPE
        assert (
            numpy.diff(time.astype(numpy.float)) > 0
        ).all(), "times must be strictly monotonically increasing"
        assert isinstance(fs, int)
        assert fs > 0
        # assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        assert not (len(time) and duration <= time[-1]), "duration is not > times"
        self._fs = fs
        self._time = time
        self._duration = TIME_TYPE(duration)

    def get_time(self):
        assert (
            numpy.diff(self._time.astype(numpy.float)) > 0
        ).all(), "times must be strictly monotonically increasing"
        # in case the user messed with .time[index] directly
        return self._time

    def set_time(self, time):
        assert isinstance(time, numpy.ndarray)
        assert time.ndim == 1
        assert time.dtype == TIME_TYPE
        assert (
            numpy.diff(time.astype(numpy.float)) > 0
        ).all(), "times must be strictly monotonically increasing"
        # assert (numpy.diff(time.astype(numpy.float)) >= 0).all(), "times must be strictly monotonically increasing"
        assert not (len(time) and self._duration <= time[-1]), "duration is not > times"
        self._time = time

    def get_value(self):
        raise Exception("No values exist for Events")

    def set_value(self, value):
        raise Exception("can't set values for Events")

    def get_duration(self):
        return self._duration

    def set_duration(self, duration):
        assert isinstance(duration, TIME_TYPE) or isinstance(duration, int)
        assert not (
            len(self._time) and duration <= self._time[-1]
        ), "duration is not > times"
        self._duration = duration

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    time = property(get_time, set_time)
    value = property(get_value, set_value)
    duration = property(get_duration, set_duration, doc="duration of track")
    fs = property(get_fs, set_fs, doc="sampling frequency")

    def __len__(self):
        return len(self._time)

    def __str__(self):
        return "%s, fs=%i, duration=%i." % (self._time, self.fs, self.duration)

    def __add__(self, other):
        assert type(other) == type(self), "Cannot add Track objects of different types"
        assert self.fs == other.fs, "sampling frequencies must match"
        time = numpy.concatenate(
            (self.time, (other.time + self.duration).astype(other.time.dtype))
        )
        duration = self.duration + other.duration
        return type(self)(time, self.fs, duration)

    def __eq__(self, other):
        if (
            (self._fs == other._fs)
            and (self._duration == other._duration)
            and (len(self._time) == len(other._time))
            and (self._time == other._time).all()
        ):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def crossfade(self, event, length):
        """append wave to self, using a crossfade of a specified length in samples"""
        assert type(self) == type(event), "Cannot add Track objects of different types"
        assert self.fs == event.fs, "sampling frequency of waves must match"
        assert isinstance(length, int)
        assert length > 0
        assert event.duration >= length
        assert self.duration >= length
        # cut left, cut right, concatenate
        a = self.select(0, self.duration - length // 2)
        b = event.select(length - length // 2, event.duration)
        return a + b

    def resample(self, fs):
        if fs != self._fs:
            factor = fs / self._fs
            # need to use numpy.round for consistency - it's different from the built-in round
            duration = int(numpy.ceil(factor * self._duration))
            if len(self._time):
                time = numpy.round(factor * self._time).astype(TIME_TYPE)
                if (numpy.diff(time) == 0).any():
                    logger.warning(
                        "new fs causes times to fold onto themselves due to lack in precision, "
                        "eliminating duplicates"
                    )
                    time = numpy.unique(time)
                if duration <= time[-1]:  # try to fix this situation
                    if len(time) > 1:
                        if (
                            time[-2] == time[-1] - 1
                        ):  # is the penultimate point far enough away?
                            raise Exception(
                                "cannot adjust last time point to be smaller "
                                "than the duration of the track"
                            )
                    logger.warning(
                        "new fs causes last time point to be == duration, "
                        "retarding last time point by one sample"
                    )
                    time[-1] -= 1
            else:
                time = self._time
            return type(self)(time, fs, duration)
        else:
            return self

    def select(self, a, b):
        assert a >= 0
        assert b > a
        assert b <= self._duration
        ai = self.time.searchsorted(a)
        bi = self.time.searchsorted(b)
        time = self._time[ai:bi] - a
        return type(self)(time, self.fs, b - a)

    default_suffix = ".evt"

    @classmethod
    def read(cls, name, fs, duration, *_args, **_kwargs):
        """Loads object from name, adding default extension if missing."""
        if name == os.path.splitext(name)[0]:
            ext = cls.default_suffix
            name += ext
        else:
            ext = os.path.splitext(name)[1].lower()
        if ext == ".pml" or ext == cls.default_suffix:
            self = cls.read_pml(name, fs)
        elif ext == ".pp":
            self = cls.read_PointProcess(name, fs)
        else:
            raise ValueError("file '{}' has unknown format".format(name))
        if duration:
            self.set_duration(duration)
        return self

    @classmethod
    def read_pml(cls, name, fs=48000):
        with open(name, "r") as f:
            lines = f.readlines()
        if len(lines) == 0:
            logger.warning("pmlread(): empty file")
            return Event(numpy.empty(0, dtype=TIME_TYPE), fs, 0)
        time = numpy.zeros(len(lines), TIME_TYPE)
        for i, line in enumerate(lines):
            token = line.split(" ")
            # t1 = token[0]
            t2 = token[1]
            time[i] = numpy.round(float(t2) * fs)
        if (numpy.diff(time) <= 0).any():
            logger.error(
                "events are too close (for fs=%i) in file: %s, merging events"
                % (fs, name)
            )
            time = numpy.unique(time)
        # we cannot truly know the duration, so we are giving it the minimum duration
        return Event(time, fs, int(time[-1] + 1))

    pmlread = read_pml

    @classmethod
    def read_pm(cls, name, fs, _duration):
        # suited for loading .pm files (pitch mark) that exist in CMU-ARCTIC
        with open(name, "r") as f:
            lines = f.readlines()
        if len(lines) == 0:
            logger.warning("pmread(): empty file")
            return Event(numpy.empty(0, dtype=TIME_TYPE), fs, 0)
        time = numpy.zeros(len(lines), TIME_TYPE) - 1
        for i, line in enumerate(lines):
            token = line.split(" ")
            t1 = token[0]
            with contextlib.suppress(IndexError):
                time[i] = numpy.round(float(t1) * fs)
            # try:
            #     time[i] = numpy.round(float(t1) * fs)
            # except IndexError:
            #     continue
            # else:
            #     t2 = token[1]

        time = time[time != -1]
        if (numpy.diff(time) <= 0).any():
            logger.error(
                "events are too close (for fs=%i) in file: %s, merging events"
                % (fs, name)
            )
            time = numpy.unique(time)
        # if int(time[-1] + 1) >= duration:
        # time = time[:-1]
        # we cannot truly know the duration, so we are giving it the minimum duration
        return Event(time, fs, int(time[-1] + 1))

    def write(self, name, *_args, **_kwargs):
        """Saves object to name, adding default extension if missing."""
        name_wo_ext = os.path.splitext(name)[0]
        if name == name_wo_ext:
            name += self.default_suffix
        self.write_pml(name)

    def write_pml(self, name):
        f = open(name, "w")
        t1 = 0.0
        for t in self.time:
            t2 = t / self.fs
            f.write("%f %f .\n" % (t1, t2))
            t1 = t2
        f.close()

    pmlwrite = write_pml

    # def __getitem__(self, index):
    #     return self._time[index]

    # def __setitem__(self, index, value):
    #     self._time[index] = value

    def get(self, t):
        if t in self._time:
            return True
        else:
            return False

    def draw_pg(self, **kwargs):
        raise NotImplementedError

    def time_warp(self, X, Y):
        assert X[0] == 0
        assert Y[0] == 0
        assert X[-1] == self.duration
        time = numpy.interp(self.time, X, Y).astype(self.time.dtype)
        if 0:
            # from matplotlib import pylab
            # pylab.plot(X, Y, 'rx-')
            # for x, y in zip(self.time, time):
            #     pylab.plot([x, x, 0], [0, y, y])
            # pylab.show()
            raise NotImplementedError
        # may have to remove some collapsed items
        assert len(numpy.where(numpy.diff(time) == 0)[0]) == 0
        self._time = time  # [index]
        self.duration = Y[-1]  # changed this from _duration

    #  TODO: NEEDS TESTING!
    def insert(self, a, t):  #
        assert isinstance(t, type(self))
        index = numpy.where((self.time >= a))[0][0]  # first index of the "right side"
        self._time = numpy.hstack(
            (self._time[:index], t._time + a, self._time[index:] + t.duration)
        )
        self._duration += t.duration



# -*- coding: utf-8 -*-
from typing import NamedTuple, Optional, Tuple

import numpy as np
from scipy import signal
from signalworks import dsp
from signalworks.processors.processing import DefaultProgressTracker, Processor
from signalworks.tracking import TimeValue, Wave


class EnergyEstimator(Processor):
    name = "RMS-Energy (dB)"
    acquire = NamedTuple("acquire", [("wave", Wave)])

    def __init__(self):
        super().__init__()
        self.parameters = {
            "frame_size": 0.020,  # in seconds
            "frame_rate": 0.010,
        }  # in seconds

    def process(
        self, progressTracker: Optional[DefaultProgressTracker] = None
    ) -> Tuple[TimeValue]:
        # Processor.process(self, **kwargs)
        if progressTracker is not None:
            self.progressTracker = progressTracker
        wav = self.data.wave
        assert isinstance(wav, Wave)
        wav = wav.convert_dtype(np.float64)
        self.progressTracker.update(10)
        assert isinstance(wav, Wave)
        frame = dsp.frame(
            wav, self.parameters["frame_size"], self.parameters["frame_rate"]
        )
        self.progressTracker.update(70)
        frame.value *= signal.hann(frame.value.shape[1])
        value = 20 * np.log10(np.mean(frame.value ** 2.0, axis=1) ** 0.5)
        self.progressTracker.update(90)
        nrg = TimeValue(
            frame.time,
            value,
            wav.fs,
            wav.duration,
            path=wav.path.with_name(wav.path.stem + "-energy").with_suffix(
                TimeValue.default_suffix
            ),
        )
        nrg.min = value.min()
        nrg.max = value.max()
        nrg.unit = "dB"
        return (nrg,)



# -*- coding: utf-8 -*-
"""Viterbi"""

import numpy as np
import numba


# eager compilation through function signature
@numba.jit((numba.float64[:, :], numba.float64), nopython=True, cache=True)
def search_smooth(ftr, smooth):  # 100% function-based
    """P is the number of candidates at each time
    T is the number of available time points"""
    T, P = ftr.shape
    assert T > 1
    assert P < 65536  # could adjust dtype automatically, currently set at uint16
    trans = np.empty(P)
    zeta = np.empty(P)
    # one row for previous, and one for current score (no tracking for all t, saving memory)
    score = np.empty((2, P))
    # often this matrix has less than %1 of meaningful entries, could be made sparse
    path = np.zeros((T, P), np.uint16) - 1
    # the - 1 (really: 65535) helps catch bugs
    seq = np.zeros(T, np.uint16) - 1
    # forward
    for t in range(T):
        # print(t, T)
        current = t % 2  # 2-cell ring buffer pointer
        previous = (t - 1) % 2
        # OBSERVATION t
        # observe(t, ftr, score[score_cur])  # score is the output
        score[current, :] = ftr[t]
        # OBSERVATION t
        if t == 0:
            jindex = np.where(score[0])[0]  # active FROM nodes
            assert len(jindex), "no observations for target[0]"
        else:
            iindex = np.where(score[current] > -np.inf)[0]  # possible TO nodes
            # assert len(iindex), 'no observation probabilities above pruning threshold for target[%d]' % t
            for i in iindex:  # TO this node - TODO: this may be parallelizable
                # TRANSITION jindex -> i @ t
                # transition(t, ftr, jindex, i, trans)  # trans is the output
                trans[jindex] = -np.abs(jindex - i) * smooth
                # TRANSITION jindex -> i @ t
                # zeta[jindex] = score[score_prv, jindex] + trans[jindex]
                # zeta[jindex] = score[score_prv][jindex] + trans[jindex]
                zeta = (
                    score[previous] + trans
                )  # really only needed over jindex, but indexing is slow
                path[t, i] = zindex = jindex[zeta[jindex].argmax()]
                score[current, i] += zeta[zindex]
            assert np.any(
                score[current] > -np.inf
            ), "score/prob[t] must not be all below pruning threshold"
            jindex = iindex  # new active FROM nodes
    # backward
    assert current == (T - 1) % 2
    # seq[-1] = iindex[score[score_cur, iindex].argmax()]
    seq[-1] = iindex[score[current][iindex].argmax()]
    for t in range(T - 1, 0, -1):
        seq[t - 1] = path[t, seq[t]]
    return seq, score[current, seq[-1]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
from setuptools import setup, find_packages
from typing import List


def read(fname: str) -> List[str]:
    contents = open(os.path.join(os.path.dirname(__file__), fname)).read()
    return contents.split("\n")


def read_requirements(fname: str) -> List[str]:
    contents = read(fname)[1:]
    filtered = []
    for requirement in contents:
        if requirement.startswith("git+"):
            _, _, pkg_name = requirement.rpartition("=")
            requirement = pkg_name + "@" + requirement
        elif requirement.startswith("-i"):
            continue
        elif requirement.endswith("."):
            continue
    filtered.append(requirement)
    return filtered


with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

setup(
    author="Alex Kain",
    author_email="lxkain@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Library to handle signal data and perform signal processing computations",
    entry_points={"console_scripts": ["signalworks=signalworks.cli:main"]},
    install_requires=["numpy", "scipy", "numba"],
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="signalworks",
    name="signalworks",
    packages=find_packages(),
    setup_requires=["pytest-runner"],
    test_suite="tests",
    url="https://github.com/lxkain/signalworks",
    version="0.1.2",
    zip_safe=False,
)



# -*- coding: utf-8 -*-
"""Tracks
Each track has a fs and a duration. There are 4 kinds of tracks:

1 Event - times
2 Wave - values
3 TimeValue - values at times, duration
4 Partition - values between times

All track intervals are of the type [), and duration points to the next unoccupied sample == length
"""

import logging
from builtins import str
from pathlib import Path
from typing import List

import numpy
from signalworks.tracking.metatrack import MetaTrack
from signalworks.tracking.partition import Partition
from signalworks.tracking.timevalue import TimeValue
from signalworks.tracking.wave import Wave

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
# logger.setLevel(logging.WARNING)
# logger.setLevel(logging.ERROR)

TIME_TYPE = numpy.int64


def convert_dtype(source, target_dtype):
    """
    return a link (if unchanged) or copy of signal in the specified dtype (often changes bit-depth as well)
    """
    assert isinstance(source, numpy.ndarray)
    source_dtype = source.dtype
    assert source_dtype in (
        numpy.int16,
        numpy.int32,
        numpy.float32,
        numpy.float64,
    ), "source must be a supported type"
    assert target_dtype in (
        numpy.int16,
        numpy.int32,
        numpy.float32,
        numpy.float64,
    ), "target must be a supported type"
    if source_dtype == target_dtype:
        return source
    else:  # conversion
        if source_dtype == numpy.int16:
            if target_dtype == numpy.int32:
                return source.astype(target_dtype) << 16
            else:  # target_dtype == numpy.float32 / numpy.float64:
                return source.astype(target_dtype) / (1 << 15)
        elif source_dtype == numpy.int32:
            if target_dtype == numpy.int16:
                return (source >> 16).astype(target_dtype)  # lossy
            else:  # target_dtype == numpy.float32 / numpy.float64:
                return source.astype(target_dtype) / (1 << 31)
        else:  # source_dtype == numpy.float32 / numpy.float64
            M = numpy.max(numpy.abs(source))
            limit = 1 - 1e-16
            if M > limit:
                factor = limit / M
                logger.warning(
                    f"maximum float waveform value {M} is beyond [-{limit}, {limit}],"
                    f"applying scaling of {factor}"
                )
                source *= factor
            if target_dtype == numpy.float32 or target_dtype == numpy.float64:
                return source.astype(target_dtype)
            else:
                if target_dtype == numpy.int16:
                    return (source * (1 << 15)).astype(target_dtype)  # dither?
                else:  # target_dtype == numpy.int32
                    return (source * (1 << 31)).astype(target_dtype)  # dither?


class Track(MetaTrack):
    default_suffix = ".trk"

    def __init__(self, path):
        self._fs = 0
        self.type = None
        self.min = None
        self.max = None
        self.unit = None
        self.label = None
        if path is None:
            path = str(id(self))
        self.path = Path(path).with_suffix(self.default_suffix)

    def get_time(self):
        raise NotImplementedError

    def set_time(self, time):
        raise NotImplementedError

    time = property(get_time, set_time)

    def get_value(self):
        raise NotImplementedError

    def set_value(self, value):
        raise NotImplementedError

    value = property(get_value, set_value)

    def get_fs(self):
        return self._fs

    def set_fs(self, _value):
        raise Exception("Cannot change fs, try resample()")

    fs = property(get_fs, set_fs, doc="sampling frequency")

    def get_duration(self):
        pass

    def set_duration(self, duration):
        raise NotImplementedError

    duration = property(get_duration, set_duration)

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError

    def __len__(self):
        pass

    def __str__(self):
        pass

    def __add__(self, other):
        raise NotImplementedError

    @classmethod
    def read(cls, path):
        """Loads object from name, adding default extension if missing."""
        # E = []
        suffix = Path(path).suffix
        if suffix == ".wav":
            channels = None
            mmap = False
            return Wave.wav_read(path, channels, mmap)
        elif suffix == ".tmv":
            return TimeValue.read_tmv(path)  # for now, handle nans
        elif suffix == ".lab":
            return Partition.read(path)
        else:
            raise Exception(f"I don't know how to read files with suffix {suffix}")

    def write(self, name, *args, **kwargs):
        """Saves object to name, adding default extension if missing."""
        raise NotImplementedError

    def resample(self, fs):
        """resample self to a certain fs"""
        raise NotImplementedError

    def select(self, a, b):
        """
        return a selection of the track from a to b. a and b are in fs units.
        Times are new objects, but values are views - idea is to make a read-only section, not a copy
        """
        raise NotImplementedError

    def insert(self, a, t):
        raise NotImplementedError

    def remove(self, a, b):
        raise NotImplementedError

    def copy(self, a, b):
        raise NotImplementedError

    def cut(self, a, b):
        t = self.copy(a, b)
        self.remove(a, b)
        return t


def get_track_classes() -> List[Track]:
    def all_subclasses(c):
        return c.__subclasses__() + [
            a for b in c.__subclasses__() for a in all_subclasses(b)
        ]

    return [obj for obj in all_subclasses(Track)]


# TODO: class NamedEvent(_Track)
#  there hasn't been a need for it yet, but may be useful in the future
#  wonder if I can extend Event itself with optional values...
# class NamedEvent(_Track):
#  def __init__(self, time, value, fs, duration)


# class HetMultiTrack(MultiTrack):  # may want to define common abstract class instead
#     """
#     A dictionary containing time-synchronous tracks of equal duration, but HETEROGENOUS fs
#     """

#     # this fs relates to the manner by which we time-index (possibly with float) into the multitrack object.
#     # Use 1.0 for seconds.
#     def __init__(self, mapping=dict(), fs=1.0):
#         dict.__init__(self, mapping)
#         if __debug__:  # long assert - TODO: do this on mapping, and then assign
#             self.check()
#         self._fs = fs

#     def check(self):
#         if len(self) > 1:
#             duration = None
#             for i, (key, track) in enumerate(self.items()):
#                 if duration is None:
#                     duration = track.duration / track.fs
#                 if track.duration / track.fs != duration:
#                     raise AssertionError(
#                         f"all durations must be equal, track #{i} ('{key}') does not match track #1"
#                     )

#     def get_fs(self):
#         if len(self):
#             return self._fs
#         else:
#             return 0  # or raise?

#     def set_fs(self, fs):
#         self._fs = fs

#     fs = property(get_fs, set_fs, doc="sampling frequency of time-index")

#     def select(self, a, b, keys=None):
#         assert a >= 0
#         assert a < b  # or a <= b?
#         assert b <= self.duration
#         """return a new object with all track views from time a to b"""
#         if keys is None:
#             keys = self.keys()
#         obj = type(self)()
#         for key in keys:
#             trk = self[key]
#             obj[key] = trk.select(
#                 a / self._fs * trk._fs, b / self._fs * trk._fs
#             )  # untested
#         return obj

# def test_pml(self):
#     import tempfile
#     tmp = tempfile.NamedTemporaryFile(prefix='test_pml_')
#     filename = tmp.name
#     tmp.close()
#     self.t.pmlwrite(filename)
#     s = Event.pmlread(filename)
#     os.unlink(filename)
#     # duration CANNOT be encoded in the file (or can it?)
#     s.duration = int(numpy.round(self.t.duration * s.fs / self.t.fs))
#     s = s.resample(self.t.fs)
#     self.assertTrue(numpy.allclose(s.time, self.t.time))



import numpy as np
import pytest
from signalworks.tracking.wave import Wave


@pytest.fixture
def var():
    w = Wave(value=np.arange(0, 16000), fs=16000)
    v = Wave(value=np.arange(100, 200), fs=16000)
    return w, v


def test_add(var):
    w, v = var
    t = w + v
    assert t.duration == 16100
    assert t.value[16050] == 150
    w += v
    assert w.duration == 16100
    assert w.value[16050] == 150


def test_select(var, benchmark):
    w, _ = var
    w = benchmark(w.select, 10, 20)
    assert w == Wave(np.arange(10, 20, dtype=np.int64), 16000)


def test_resample(benchmark):
    w1 = Wave(value=np.arange(0, 50), fs=1)
    w2 = benchmark(w1.resample, 2)
    assert w1.value[0] == int(w2.value[0])


def test_crossfade(benchmark):
    wav1 = Wave(np.array([1, 1, 1, 1, 1]), 1)
    wav2 = Wave(np.array([10, 10, 10, 10, 10]), 1)
    length = 3
    wav = benchmark(wav1.crossfade, wav2, length)
    assert wav1.duration + wav2.duration - length, wav.duration
    assert np.allclose(wav.value, np.array([1, 1, 3, 5, 7, 10, 10]))



# -*- coding: utf-8 -*-
from .metatrack import MetaTrack  # noqa
from .error import LabreadError, MultiChannelError  # noqa
from .event import Event  # noqa
from .label import Label  # noqa
from .partition import Partition  # noqa
from .timevalue import TimeValue  # noqa
from .tracking import Track  # noqa: F401
from .value import Value  # noqa
from .wave import Wave  # noqa
from .multitrack import MultiTrack  # noqa



from collections import UserDict

import numpy as np
import pytest
from signalworks.tracking import Event, MultiTrack, Partition, Wave


@pytest.fixture
def var():
    w = Wave(value=np.arange(0, 50), fs=1)
    p = Partition(
        np.array([0, 15, 20, 50], dtype=np.int64),
        np.array(["start", "middle", "end"]),
        1,
    )
    e = Event(np.array([0, 10, 30]).astype(np.int64), 1, 50)
    return w, p, e


def test_select(var, benchmark):
    dict: UserDict = UserDict()
    dict["wave"], dict["partition"], dict["event"] = var
    mt = MultiTrack(dict)
    new_mt = benchmark(mt.select, 10, 24)
    assert new_mt["wave"] == Wave(value=np.arange(10, 24), fs=1)
    assert new_mt["partition"] == Partition(
        np.array([0, 5, 10, 14], dtype=np.int64),
        np.array(["start", "middle", "end"]),
        1,
    )


def test_crossfade(var, benchmark):
    w1, p1, e1 = var
    dict: UserDict = UserDict()
    dict["wave"], dict["partition"] = w1, p1
    mt1 = MultiTrack(dict)

    w2 = Wave(value=np.arange(0, 50), fs=1)
    p2 = Partition(
        np.array([0, 15, 20, 50], dtype=np.int64),
        np.array(["start", "middle", "end"]),
        1,
    )
    dict = UserDict()
    dict["wave"], dict["partition"] = w2, p2
    mt2 = MultiTrack(dict)

    mt3 = benchmark(mt1.crossfade, mt2, 5)
    assert mt3.duration == mt1.duration + mt2.duration - 5
    assert mt3["wave"] == Wave(
        value=np.r_[np.arange(45), np.array([37, 31, 24, 18, 11]), np.arange(5, 50)],
        fs=1,
    )
    assert mt3["partition"] == Partition(
        np.array([0, 15, 20, 48, 60, 65, 95], dtype=np.int64),
        np.array(["start", "middle", "end", "start", "middle", "end"]),
        1,
    )
